/*
 * AuthenticationPanelTest.java
 * JUnit based test
 *
 * Created on 3. Juli 2006, 23:15
 */
/*
 * Copyright 2006 Schlichtherle IT Services
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.schlichtherle.key.passwd.swing;

import java.awt.*;
import java.lang.ref.*;
import java.io.*;
import java.util.logging.*;

import javax.swing.*;

import junit.framework.*;

import org.netbeans.jemmy.*;
import org.netbeans.jemmy.operators.*;
import org.netbeans.jemmy.util.*;

/**
 * @author Christian Schlichtherle
 * @version @version@
 * @since TrueZIP 6.1
 */
public class AuthenticationPanelTest extends TestCase {
    static {
        JemmyProperties.setCurrentOutput(TestOut.getNullOutput()); // shut up!
    }

    private static final Logger logger
            = Logger.getLogger(AuthenticationPanelTest.class.getName());
    
    static final File rootDir;
    static {
        if (File.separatorChar == '\\')
            rootDir = new File("C:\\");
        else
            rootDir = new File("/");
    }
    
    private final ComponentChooser keyFileChooser
                = new NameComponentChooser("keyFileChooser");

    public AuthenticationPanelTest(String testName) {
        super(testName);
    }

    /**
     * Test of getFileChooser method, of class de.schlichtherle.key.passwd.swing.AuthenticationPanel.
     */
    public void testGetFileChooser() throws InterruptedException {
        final ReferenceQueue queue = new ReferenceQueue();
        final PhantomReference ref = initQueue(queue, rootDir);

        byte[] bfo;
        int i = 1;
        while (null == queue.poll()) {
            // Allocate big fat object in order to cause the internal
            // cache for the file chooser to be cleared.
            try {
                bfo = new byte[i * 1024 * 1024];
                i++; // is not reached on OOME!
            } catch (OutOfMemoryError oome) {
                //oome.printStackTrace();
                // SoftReferences (which are used by getFileChooser) are
                // guaranteed to be cleared BEFORE an OOME is thrown.
                // However, the additional chunk of memory we are requesting
                // may be too big, in which case the reference is first cleared
                // and then immediately an OOME is thrown.
                // When the SoftReference has been cleared, the referent is
                // made eligible for finalization.
                // Run the finalization now and assert this:
                System.runFinalization();
                assertNotNull(queue.poll()); // JFileChooser has been finalized
                break;
            }

            // Release memory again in order to allow the JVM to operate
            // normally.
            bfo = null;

            System.gc();
        }
        assert null == queue.poll(); // previous poll() has removed the reference
        logger.fine("Successfully allocated " + i + " megabytes before JFileChooser was discarded.");
        
        // Now ask for a file chooser again.
        JFileChooser fc = AuthenticationPanel.getFileChooser();
        assertEquals(
                "Newly instantiated JFileChooser needs to have same current directory as previous instance!",
                rootDir, fc.getCurrentDirectory());
    }

    private PhantomReference initQueue(
            final ReferenceQueue queue,
            final File dir) {
        JFileChooser fc = AuthenticationPanel.getFileChooser();
        fc.setCurrentDirectory(dir);

        final PhantomReference ref = new PhantomReference(fc, queue);

        fc = null;
        System.gc();

        fc = AuthenticationPanel.getFileChooser();
        fc.setCurrentDirectory(dir);

        fc = null;
        System.gc();

        assertNull(
                "Initial JFileChooser instance should not yet have been thrown away!",
                queue.poll());

        return ref;
    }

    public void testTabbedPane() {
        final String text = "Hello world!";
        EventQueue.invokeLater(new Runnable() {
            public void run() {
                final AuthenticationPanel instance = new AuthenticationPanel();

                JPanel passwdPanel = null;
                try {
                    instance.setPasswdPanel(passwdPanel);
                    fail("Calling setPasswdPanel(null) should throw an NPE!");
                } catch (NullPointerException npe) {
                }

                passwdPanel = new JPanel();
                passwdPanel.add(new JLabel(text));
                instance.setPasswdPanel(passwdPanel);

                final JFrame frame = new JFrame();
                frame.getContentPane().add(instance);
                frame.pack();
                frame.setLocationRelativeTo(null);
                frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
                frame.setVisible(true);
            }
        });

        final JFrameOperator frame = new JFrameOperator();

        new JTabbedPaneOperator(frame).selectPage(AuthenticationPanel.AUTH_KEY_FILE); // select tab for key files
        new JButtonOperator(frame, keyFileChooser).push(); // open file chooser
        JFileChooserOperator fc = new JFileChooserOperator();
        assertSame(AuthenticationPanel.getFileChooser(), fc.getSource());
        fc.cancel();
        new JTabbedPaneOperator(frame).selectPage(AuthenticationPanel.AUTH_PASSWD); // select tab for passwords
        new JLabelOperator(frame, text);
        fc = null;

        new JTabbedPaneOperator(frame).selectPage(AuthenticationPanel.AUTH_KEY_FILE); // select tab for key files
        new JButtonOperator(frame, keyFileChooser).push(); // open file chooser
        fc = new JFileChooserOperator();
        final File file = new File(rootDir, "test");
        fc.setSelectedFile(file);
        fc.approve();
        JTextFieldOperator tfOp = new JTextFieldOperator(frame);
        assertEquals(file.getPath(), tfOp.getText());
        fc = null;
        
        new JTabbedPaneOperator(frame).selectPage(AuthenticationPanel.AUTH_PASSWD); // select tab for passwords
        new JLabelOperator(frame, text);
        
        //frame.close();
    }
}
