/*
 * ZipEntryTest.java
 * JUnit based test
 *
 * Created on 4. Juli 2005, 15:40
 */
/*
 * Copyright 2005-2007 Schlichtherle IT Services
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.schlichtherle.util.zip;

import java.text.*;
import java.util.*;

import junit.framework.*;

/**
 * @author Christian Schlichtherle
 */
public class ZipEntryTest extends TestCase {
    
    private ZipEntry entry;
    
    public ZipEntryTest(String testName) {
        super(testName);
    }

    protected void setUp() throws Exception {
        entry = new ZipEntry("test");
    }

    protected void tearDown() throws Exception {
        entry = null;
    }

    public void testName() {
        assertEquals("test", entry.getName());
        try {
            entry.setName("foo");
            fail("Excpected IllegalStateException!");
        } catch (IllegalStateException ex) {
        }

        entry = new ZipEntry(entry); // unlock name in copy
        assertEquals("test", entry.getName());
        entry.setName("foo");
        try {
            entry.setName("bar");
            fail("Excpected IllegalStateException!");
        } catch (IllegalStateException ex) {
        }
        assertEquals("foo", entry.getName());

        entry = (ZipEntry) entry.clone(); // unlock name in clone
        assertEquals("foo", entry.getName());
        entry.setName("bar");
        try {
            entry.setName("foobar");
            fail("Excpected IllegalStateException!");
        } catch (IllegalStateException ex) {
        }
        assertEquals("bar", entry.getName());
    }

    public void testPlatform() {
        assertEquals(ZipEntry.UNKNOWN, entry.getPlatform());
        entry.setPlatform((short) 0x00);
        assertEquals((short) 0x00, entry.getPlatform());
        entry.setPlatform(ZipEntry.PLATFORM_FAT);
        assertEquals(ZipEntry.PLATFORM_FAT, entry.getPlatform());
        entry.setPlatform(ZipEntry.PLATFORM_UNIX);
        assertEquals(ZipEntry.PLATFORM_UNIX, entry.getPlatform());
        entry.setPlatform((short) 0xff);
        assertEquals((short) 0xff, entry.getPlatform());
        entry.setPlatform(ZipEntry.UNKNOWN);
        assertEquals(ZipEntry.UNKNOWN, entry.getPlatform());

        try {
            entry.setPlatform((short) (ZipEntry.UNKNOWN - 1));
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        try {
            entry.setPlatform((short) (0xff + 1));
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }
    }

    public void testGeneral() {
        assertEquals(ZipEntry.UNKNOWN, entry.getGeneral());
        entry.setGeneral(0);
        assertEquals(0, entry.getGeneral());
        entry.setGeneral(8);
        assertEquals(8, entry.getGeneral());
        entry.setGeneral(1 << 11);
        assertEquals(1 << 11, entry.getGeneral());
        entry.setGeneral(0xffff);
        assertEquals(0xffff, entry.getGeneral());
        entry.setGeneral(ZipEntry.UNKNOWN);
        assertEquals(ZipEntry.UNKNOWN, entry.getGeneral());

        try {
            entry.setGeneral(ZipEntry.UNKNOWN - 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        try {
            entry.setGeneral(0xffff + 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }
    }

    public void testGeneralBit() {
        for (int i = -1; i < 17; i++) {
            try {
                entry.getGeneralBit(i);
                fail("Excepted IllegalStateException");
            } catch (IllegalStateException ex) {
            }
        }

        try {
            entry.setGeneralBit(-1, false);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        for (int i = 0; i < 16; i++) {
            entry.setGeneralBit(i, false);
            assertFalse(entry.getGeneralBit(i));
            entry.setGeneralBit(i, true);
            assertTrue(entry.getGeneralBit(i));
        }

        try {
            entry.setGeneralBit(16, false);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        entry.setGeneral(ZipEntry.UNKNOWN);
        try {
            entry.getGeneralBit(0);
            fail("Excepted IllegalStateException");
        } catch (IllegalStateException ex) {
        }
    }

    public void testMethod() {
        assertEquals(ZipEntry.UNKNOWN, entry.getMethod());
        entry.setMethod(0);
        assertEquals(0, entry.getMethod());
        entry.setMethod(ZipEntry.STORED);
        assertEquals(ZipEntry.STORED, entry.getMethod());
        entry.setMethod(ZipEntry.DEFLATED);
        assertEquals(ZipEntry.DEFLATED, entry.getMethod());
        entry.setMethod(0xffff);
        assertEquals(0xffff, entry.getMethod());
        entry.setMethod(ZipEntry.UNKNOWN);
        assertEquals(ZipEntry.UNKNOWN, entry.getMethod());

        try {
            entry.setMethod(ZipEntry.UNKNOWN - 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        try {
            entry.setMethod(0xffff + 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }
    }

    public void testDosTime() {
        assertEquals(ZipEntry.UNKNOWN, entry.getDosTime());
        entry.setDosTime(0);
        assertEquals(0, entry.getDosTime());
        entry.setDosTime(0xffffffffL);
        assertEquals(0xffffffffL, entry.getDosTime());
        entry.setDosTime(ZipEntry.UNKNOWN);
        assertEquals(ZipEntry.UNKNOWN, entry.getDosTime());

        try {
            entry.setDosTime(ZipEntry.UNKNOWN - 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        try {
            entry.setDosTime(0xffffffffL + 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }
    }

    public void testTimeConversion() {
        long dosTime = ZipEntry.java2dosTime(0);
        assertEquals(ZipEntry.MIN_DOS_TIME, dosTime);
        assertEquals(dosTime, ZipEntry.java2dosTime(ZipEntry.dos2javaTime(dosTime)));
        dosTime = ZipEntry.java2dosTime(System.currentTimeMillis());
        assertEquals(dosTime, ZipEntry.java2dosTime(ZipEntry.dos2javaTime(dosTime)));

        try {
            ZipEntry.dos2javaTime(-1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        try {
            ZipEntry.dos2javaTime(0xffffffffL + 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        try {
            ZipEntry.java2dosTime(-1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }
}

    public void testTime() {
        assertEquals(ZipEntry.UNKNOWN, entry.getTime());
        entry.setTime(0);
        long time = entry.getTime();
        assertEquals(ZipEntry.dos2javaTime(ZipEntry.MIN_DOS_TIME), time);
        assertEquals("01.01.1980 00:00:00 CET",
                new MessageFormat("{0,date} {0,time,long}", Locale.GERMANY)
                .format(new Object[] { new Date(time) }));
        entry.setTime(ZipEntry.UNKNOWN);
        assertEquals(ZipEntry.UNKNOWN, entry.getTime());

        try {
            entry.setTime(ZipEntry.UNKNOWN - 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }
    }

    public void testCrc() {
        assertEquals(ZipEntry.UNKNOWN, entry.getCrc());
        entry.setCrc(0);
        assertEquals(0, entry.getCrc());
        entry.setCrc(0xffffffffL);
        assertEquals(0xffffffffL, entry.getCrc());
        entry.setCrc(ZipEntry.UNKNOWN);
        assertEquals(ZipEntry.UNKNOWN, entry.getCrc());

        try {
            entry.setCrc(ZipEntry.UNKNOWN - 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        try {
            entry.setCrc(0xffffffffL + 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }
    }

    public void testCompressedSize() {
        assertEquals(ZipEntry.UNKNOWN, entry.getCompressedSize());
        entry.setCompressedSize(0);
        assertEquals(0, entry.getCompressedSize());
        entry.setCompressedSize(0xffffffffL);
        assertEquals(0xffffffffL, entry.getCompressedSize());
        entry.setCompressedSize(ZipEntry.UNKNOWN);
        assertEquals(ZipEntry.UNKNOWN, entry.getCompressedSize());

        try {
            entry.setCompressedSize(ZipEntry.UNKNOWN - 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        try {
            entry.setCompressedSize(0xffffffffL + 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }
    }

    public void testSize() {
        assertEquals(ZipEntry.UNKNOWN, entry.getSize());
        entry.setSize(0);
        assertEquals(0, entry.getSize());
        entry.setSize(0xffffffffL);
        assertEquals(0xffffffffL, entry.getSize());
        entry.setSize(ZipEntry.UNKNOWN);
        assertEquals(ZipEntry.UNKNOWN, entry.getSize());

        try {
            entry.setSize(ZipEntry.UNKNOWN - 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }

        try {
            entry.setSize(0xffffffffL + 1);
            fail("Excepted IllegalArgumentException");
        } catch (IllegalArgumentException ex) {
        }
    }

    public void testExtra() {
        assertEquals(null, entry.getExtra());

        final byte[] set = new byte[10];
        entry.setExtra(set);

        final byte[] got1 = entry.getExtra();
        assertNotNull(got1);
        assertNotSame(set, got1);
        assertTrue(Arrays.equals(set, got1));

        final byte[] got2 = entry.getExtra();
        assertNotNull(got2);
        assertNotSame(got1, got2);
        assertTrue(Arrays.equals(got1, got2));
    }
}
